/*
 * $Id: kextendedsslsocket.cpp,v 1.15 2004/03/07 15:16:07 blackmor Exp $
 *
 * This file is part of the kwirc project
 * Copyright (C) 2003 Julien Perrot <blackmore@no-log.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * See the file "COPYING" for the exact licensing terms.
 */

#include "config.h"
#ifdef HAVE_SSL

#include "kextendedsslsocket.h"
#include <kdebug.h>
#include "kwircdebug.h"
#include <errno.h>

bool KExtendedSSLSocket::init_done = false;
SSL_CTX * KExtendedSSLSocket::context;

KExtendedSSLSocket::KExtendedSSLSocket(const QString& host, int port,
		int flags)
	:KExtendedSocket(host, port, flags), ssl(NULL), cert(NULL),
			rbio(NULL), wbio(NULL)
{
	initSSL();

	QObject::connect(this, SIGNAL(outgoingSSLDataReady()),
			SLOT(slotSSL_outgoingSSLDataReady()));
	QObject::connect(this, SIGNAL(ssl_readyRead()),
			SLOT(slotSSL_readyRead()) );
	QObject::connect(this, SIGNAL(bytesWritten (int)),
			SLOT(slotBytesWritten(int)));
}

KExtendedSSLSocket::~KExtendedSSLSocket()
{
	if (ssl) {
		// SSL_free takes care to free the associated items
		// to the SSL object, that is to say the read and write
		// BIOs so there is no need to free them explicitly
		// (would lead to a double free which is baaaaaaaad)
		SSL_free(ssl);
		rbio = NULL;
		wbio = NULL;
	}

	if (cert)
		X509_free(cert);
}

void KExtendedSSLSocket::initSSL()
{
	if (!init_done) {
		SSL_METHOD *method;
		SSL_library_init();
		SSL_load_error_strings();
		method = SSLv23_client_method();
		context = SSL_CTX_new(method);
		init_done = true;
	}

	ssl = SSL_new(context);

	rbio = BIO_new(BIO_s_mem());
	wbio = BIO_new(BIO_s_mem());

	SSL_set_bio(ssl, rbio, wbio);
}

void KExtendedSSLSocket::slot_readyRead()
{
	int size = bytesAvailable();
	char * buf = (char *) malloc(size*sizeof(char));

	KExtendedSocket::readBlock(buf, size);
	BIO_write(rbio, buf, size);
	sslUpdate();

	free(buf);
}

void KExtendedSSLSocket::slotSSL_readyRead()
{
	emit readyRead();
}

bool KExtendedSSLSocket::isRecvData()
{
	return (bytesAvailable() > 0);
}

bool KExtendedSSLSocket::isOutgoingSSLData()
{
	return (BIO_pending(wbio) > 0);
}

void KExtendedSSLSocket::slotSSL_outgoingSSLDataReady()
{
	int ret;
	int size = BIO_pending(wbio);
	char * buffer = (char *) malloc(size*sizeof(char));
	int r = BIO_read(wbio, buffer, size);

	ret = KExtendedSocket::writeBlock(buffer, r);

	free(buffer);
}

int KExtendedSSLSocket::doConnect()
{
	int ret = SSL_connect(ssl);
	if(ret < 0) {
		kdError(SSL_DBG) << "Error in SSL_Connect" << endl;
		int x = SSL_get_error(ssl, ret);
		if(x == SSL_ERROR_WANT_CONNECT || x == SSL_ERROR_WANT_READ
				|| x == SSL_ERROR_WANT_WRITE) {
			kdWarning(SSL_DBG) << "Try again" << endl;
			return TryAgain;
		} else
			return Error;
	} else if(ret == 0)
		return Error;

	return Success;
}

int KExtendedSSLSocket::doHandshake()
{
	if ( ssl == NULL ) {
		kdError( SSL_DBG ) << "ssl is NULL" << endl;
	}

	int ret = SSL_do_handshake(ssl);
	if(ret < 0) {
		kdError(SSL_DBG) << "Error in SSL_do_handshake" << endl;
		int x = SSL_get_error(ssl, ret);
		if(x == SSL_ERROR_WANT_READ || x == SSL_ERROR_WANT_WRITE) {
			kdWarning(SSL_DBG) << "Try again" << endl;
			return TryAgain;
		}
		else
			return Error;
	}
	else if(ret == 0)
		return Error;

	if (!(cert = SSL_get_peer_certificate(ssl))) {
		kdError(SSL_DBG) << "SSL server supplied no certificate"
			<< endl;
		return Error;
	}

	switch(SSL_get_verify_result(ssl)) {
		case X509_V_OK:
			kdDebug(SSL_DBG) << "Certificate verification" <<
				" successfull" << endl;
			break;
		default:
			kdWarning(SSL_DBG) << "Certificate verification failed"
				<< endl;
			break;
	}

	return Success;
}

void KExtendedSSLSocket::sslUpdate()
{
	if(mode == Idle)
		return;

	if(mode == Connect) {
		int ret = doConnect();
		if(ret == Success) {
			mode = Handshake;
		}
		else if(ret == Error) {
			emit connectionFailed (errno);
			return;
		}
	}

	if(mode == Handshake) {
		int ret = doHandshake();
		if(ret == Success) {
			mode = Active;
			emit ssl_handshaken(true);
		}
		else if(ret == Error) {
			emit ssl_handshaken(false);
			return;
		}
	}

	if(isOutgoingSSLData()) {
		emit outgoingSSLDataReady();
	}

	// try to read incoming unencrypted data
	sslReadAll();

	if(isRecvData())
		emit ssl_readyRead();
}

unsigned KExtendedSSLSocket::feedWriteBuffer (unsigned nbytes,
		const char *buffer)
{
	int ret = SSL_write(ssl, buffer, nbytes);
	if (ret <= 0) {
		kdWarning(SSL_DBG) << "feedWriteBuffer : error in SSL_write"
			<< endl;
	}
	sslUpdate();
	return ret;
}

unsigned KExtendedSSLSocket::feedReadBuffer(unsigned nbytes,
		const char *buffer, bool)
{
	int ret = BIO_write(rbio, buffer, nbytes);
	if (ret <= 0) {
		kdWarning(SSL_DBG) << "feedReadBuffer : error in BIO_write"
			<< endl;
	}
	sslUpdate();
	return ret;
}

Q_LONG KExtendedSSLSocket::writeBlock(const char *data, Q_ULONG len)
{
	return feedWriteBuffer(len, data);
}

void KExtendedSSLSocket::sslReadAll()
{
	char b[4096];

	while(1) {
		int x = SSL_read(ssl, b, 4096);
		if(x <= 0) {
			break;
		}
		KExtendedSocket::feedReadBuffer(x, b);
	}
}

void KExtendedSSLSocket::slotBytesWritten(int)
{
	// Nothing
}

void KExtendedSSLSocket::reset()
{
	KExtendedSocket::reset();
	if (ssl) {
		SSL_free(ssl);
		rbio = NULL;
		wbio = NULL;
	}
	if (cert)
		X509_free(cert);

	initSSL();
}

int KExtendedSSLSocket::startHandshake()
{
	if (socketStatus() == connected) {
		mode = Connect;
		sslUpdate();
		return 0;
	} else {
		return 1;
	}
}

#include "kextendedsslsocket.moc"


#endif
