/*
 * $Id: kwirctheme.cpp,v 1.19 2004/03/07 15:16:07 blackmor Exp $
 *
 * This file is part of the kwirc project
 * Copyright (C) 2003 Julien Perrot
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * See the file "COPYING" for the exact licensing terms.
 */
#include "kwirctheme.h"
#include <qstring.h>
#include <qstylesheet.h>
#include <qstringlist.h>
#include <qmap.h>
#include <qregexp.h>
#include <kdebug.h>
#include "kwircdebug.h"

const char * KwircTheme::defaultFormat[NB_STRINGS][2] = {
	{"privmsg"     , "<%1> %2"},
	{"privmsg_hl"  , "%B<%1> %2%BE"},
	{"notice"      , "- %1 - %2"},
	{"notice_hl"   , "%B- %1 - %2%BE"},
	{"action"      , "* %1 %2"},
	{"action_hl"   , "%B* %1 %2%BE"},
	{"you_join"    , "--> You are now talking on %1"},
	{"join"        , "--> %1 (%2) has joined %3"},
	{"part"        , "<-- %1 has left %2 (%3)"},
	{"quit"        , "<-- %1 has quit (%2)"},
	{"nick"        , "--- %1 is now known as %2"},
	{"ctcp"        , "--- %1 CTCP %2"},
	{"invite"      , "--- %1 was invited to %2"},
	{"kick"        , "--- %1 has kicked (%2)"},
	{"ban"         , "--- %1 was added to ban list"},
	{"deban"       , "--- %1 was removed from ban list"},
	{"op"          , "--- %1 has been oped"},
	{"deop"        , "--- %1 has been deoped"},
	{"voice"       , "--- %1 has been voiced"},
	{"devoice"     , "--- %1 has been unvoiced"},
	{"newkey"      , "--- new key is %1"},
	{"delkey"      , "--- no more key"},
	{"topic"       , "--- %1 changed the topic to: %2"},
	{"kill"        , "--- %1 killed (%2)"},
	{"mode"        , "--- %1 sets mode (%2)"},
	{"whois"       , "%1"},
	{"who"         , "%1: %2"},
	{"you_act"     , "* %1 %2"},
	{"you_privmsg" , "<%1> %2"}
};

KwircTheme::KwircTheme(const QString& fileName)
{

	loadTheme(fileName);
}

KwircTheme::KwircTheme(KConfig *config)
{
	loadTheme(config);
}

void KwircTheme::setFormatString(const QString &key, const QString &formatString)
{
	m_formatStringMap[key] = formatString;
	m_computedStringMap[key] = computeString(formatString);
}


const QString &KwircTheme::getFormatString(const QString &key)
{
	return m_formatStringMap[key];
}

const QString &KwircTheme::getComputedString(const QString &key)
{
	return m_computedStringMap[key];
}

QString KwircTheme::getColoredNick(const QString &nick)
{
	uint nb_colors = m_nickColors.count();
	uint length, i =  0;
	uint total = 0;
	length = nick.length();

	if (nb_colors == 0) {
		kdWarning(THEME_DBG) << k_funcinfo << endl;
		return nick;
	}

	while(i < length) {
		total += (uint) nick[i].latin1();
		i++;
	}

	QColor color = m_nickColors[total % nb_colors];
	return QString("<font color=\"%1\">%2</font>").arg(color.name()).arg(nick);
}

/*
 * primsg_string %b<%B%1%b>%B %d%2%d
 */
QString KwircTheme::computeString(const QString &str)
{
	static const QString colorOpen = QString("<font color=\"%1\">");
	static const QString colorClose = QString("</font>");
	QString string = QStyleSheet::escape(str);		// escapes all html

	string.replace("%e", colorClose);
	string.replace("%BE", QString("</b>"));
	string.replace("%B", QString("<b>"));
	string.replace("%UE", QString("</u>"));
	string.replace("%U", QString("<u>"));

	QStringList colors = m_themeColors.keys();

	colors.sort();
	for ( int i = colors.count()-1; i >= 0; i-- ) {
		QString format = "%C" + colors[i];
		QString colorName = m_themeColors[colors[i]].name();
		string.replace(format, colorOpen.arg(colorName));
	}

	return string;
}

void KwircTheme::loadTheme(const QString &path)
{
	m_fileName = path;
	KConfig *config = new KConfig(path);
	loadTheme(config);
	delete config;
}

void KwircTheme::loadTheme(KConfig *config)
{
	QColor * defaultColor = new QColor(0, 0, 0);

	m_themeColors.clear();
	m_nickColors.clear();
	m_formatStringMap.clear();
	m_computedStringMap.clear();

	QMap<QString, QString> entries = config->entryMap("Theme");
	QStringList keys = entries.keys();
	QStringList nickColorsKeys = keys.grep(QRegExp("colorNick\\d+"));
	QStringList themeColorsKeys = keys.grep(QRegExp("color\\d+"));

	config->setGroup("Theme");

	m_timestampFormat = config->readEntry( "timestampFormat",  "hh:mm" );

	for ( QStringList::Iterator it = nickColorsKeys.begin();
	      it != nickColorsKeys.end(); ++it ) {
		m_nickColors.append(config->readColorEntry(*it,defaultColor));
	}

	for ( QStringList::Iterator it = themeColorsKeys.begin();
	      it != themeColorsKeys.end(); ++it ) {
		QRegExp re("color(\\d+)");
		re.search(*it);
		QString number = re.cap(1);
		m_themeColors[number] = config->readColorEntry(*it,defaultColor);
	}

	for (int i=0; i<NB_STRINGS; i++) {
		const char ** format = defaultFormat[i];
		QString formatString = config->readEntry(format[0],format[1]);
		m_formatStringMap[format[0]] = formatString;
	}

	updateComputedStrings();

}

void KwircTheme::setColor(const QString &key, const QColor &color)
{
	m_themeColors[key] = color;
	updateComputedStrings();
}

void KwircTheme::updateComputedStrings()
{
	QMap<QString, QString>::Iterator it;
	for (it = m_formatStringMap.begin(); it != m_formatStringMap.end(); it++) {
		m_computedStringMap[it.key()] = computeString(it.data());
	}
}

void KwircTheme::refresh()
{
	loadTheme(m_fileName);
}

const QString & KwircTheme::getTimestampFormat()
{
	return m_timestampFormat;
}

void KwircTheme::setTimestampFormat( const QString & format)
{
	m_timestampFormat = format;
}




