/*
 * $Id: nick.cpp,v 1.27 2004/03/07 15:16:07 blackmor Exp $
 *
 * This file is part of the kwirc project
 * Copyright (C) 2003 nohar
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * See the file "COPYING" for the exact licensing terms.
 */
#include "server.h"

#include <qdockwindow.h>
#include <config.h>
#include <ktoolbarbutton.h>
#include <qspinbox.h>
#include <qcheckbox.h>
#include <qlayout.h>
#include <stdio.h>
#include <stdlib.h>
#include <kfiledialog.h>
#include <kfontdialog.h>
#include <kstandarddirs.h>
#include <qpainter.h>
#include <kmenubar.h>
#include <kmessagebox.h>
#include <krootpixmap.h>
#include <kstdaction.h>
#include <klineeditdlg.h>
#include <kipc.h>
#include <dcopclient.h>
#include <klocale.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <signal.h>
#include <dirent.h>
#include <fcntl.h>
#include <assert.h>
#include <kiconloader.h>
#include <kstringhandler.h>
#include <ktip.h>
#include <netwm.h>

#include "kwirc.h"
#include <kdebug.h>
#include "kwircdebug.h"

#include "nick.h"
#include "command.h"
#include "channel.h"
#include "kwircpreferences.h"
#include "config.h"
#ifdef FEAT_RUBY
#include "rubyscriptmanager.h"
#endif

void Nick::toggleAway(const QString &awaymsg)
{
	away = !away;
	if (away)
		awayMsg = awaymsg;
	else
		awayMsg = QString::null;

	for (Channel *c = channels.first(); c; c = channels.next())
		c->updateNames();
}

void Nick::setAway(bool en)
{
	away = en;
}

bool Nick::isAway()
{
	return away;
}

bool Nick::used()
{
	return channels.count() > 0;
}

void Nick::addChannel(Channel *c)
{
	if(! channels.contains(c))
		channels.append(c);
}

void Nick::removeChannel(Channel *c)
{
	channels.remove(c);
}

void Nick::recvJoin(Channel *c)
{
	addChannel(c);
	c->recvJoin(this);
}

void Nick::recvInvite(Nick *who, Channel *where)
{
	where->recvInvite(this, who);
	//where->print(LOG_INVITE, getName(), who->getName());
}

void Nick::recvKill(const QString &what)
{
	for (Channel *c = channels.first(); c; c = channels.next() )
		c->recvKill(this, what);
	emit log(LOG_KILL, getStringId(), what, QString::null);
}

void Nick::recvKick(Nick *kicker, Channel *c, const QString &what)
{
	removeChannel(c);
	c->recvKick(this, kicker, what);
}

void Nick::recvPart(Channel *c, const QString &what)
{
	removeChannel(c);
	c->recvPart(this, what);
}

void Nick::recvQuit(const QString &what)
{
	for (Channel *c = channels.first(); c; c = channels.next() )
		c->recvQuit(this, what);

	emit log(LOG_QUIT, getStringId(), getStringId(), what);
#ifdef FEAT_RUBY
	RubyScriptManager::notifyQuit(this, what);
#endif
}

void Nick::recvPrivmsg(const QString &what)
{
	emit log(LOG_PRIVMSG, getStringId(), getStringId(), what);

#ifdef FEAT_RUBY
	RubyScriptManager::notifyPrivate(this, what);
#endif
}

#ifdef FEAT_RUBY
void Nick::recvCtcp(const QString &ctcp, const QString &rest)
#else
void Nick::recvCtcp(const QString &ctcp, const QString &)
#endif
{
	if(ctcp.compare(QString("VERSION")) == 0)
		getServer()->sendCtcpNotice(getName(),
				QString("VERSION_REPLY %1")
				.arg(KwircPreferences::getVersionReply()));
	if(ctcp.compare(QString("PING")) == 0)
		getServer()->sendCtcpNotice(getName(), QString("PONG"));
#ifdef FEAT_RUBY
	RubyScriptManager::notifyPrivateCtcp(this, ctcp, rest);
#endif
}

void Nick::recvAction(const QString &what)
{
	emit log(LOG_ACTION, getStringId(), getStringId(), what);
#ifdef FEAT_RUBY
	RubyScriptManager::notifyPrivateAction(this, what);
#endif
}

void Nick::recvNotice(const QString &what)
{
	emit log(LOG_PRIVMSG, getStringId(), getStringId(), what);
#ifdef FEAT_RUBY
	RubyScriptManager::notifyPrivateNotice(this, what);
#endif
}

void Nick::recvCtcpNotice(const QString &what)
{
	emit log(LOG_NOTICE, getStringId(), getStringId(), what);
#ifdef FEAT_RUBY
	RubyScriptManager::notifyPrivateCtcpNotice(this, what);
#endif
}

void Nick::isIn(Channel *c)
{
	if(channels.contains(c))
		return;
	channels.append(c);
}

void Nick::noMoreIn(Channel *c)
{
	channels.remove(c);
}

void Nick::setHost(const QString &ahost)
{
	host = ahost;
}

void Nick::setIdent(const QString &id)
{
	ident = id;
}

void Nick::recvNick(const QString &n)
{
	QString oldn = getStringId();
	setStringId(n);
	for (Channel *c = channels.first(); c; c = channels.next() ) {
		c->recvNick(this, oldn);
	}
	emit log(LOG_NICK, getStringId(), oldn, QString::null);
#ifdef FEAT_RUBY
	RubyScriptManager::notifyNick(this, oldn);
#endif
}

const QString &Nick::getIrcId()
{
	ircid = QString("%1!%2@%3").arg(getStringId()).arg(ident).arg(host);
	return ircid;
}

const QString &Nick::getName()
{
	return getStringId();
}

const QString &Nick::getAuxInfo()
{
	auxinfo = QString("%1@%2").arg(ident).arg(host);
	return auxinfo;
}

void Nick::setInfoString(const QString &n)
{
	if (n.contains('!')) {
		QString left = n.section('!', 1, 1);
		ident = left.section('@', 0, 0);
		host = left.section('@', 1, 1);
	}
}

//:nohar!~nohar@nucleus.estu.iie.cnam.fr PRIVMSG #37337 :yop
Nick::Nick(Server *s, const QString &n):
	Destination(s, n, TYPE_NICK),
	away(false)
{
	if (n.contains('!')) {
		setStringId(n.section('!', 0, 0));
		setInfoString(n);
	} else {
		setStringId(n);
		host = QString::null;
		ident = QString::null;
	}
#ifdef FEAT_RUBY
	setRubyInstance(RubyScriptManager::getRubyNickInstance(this));
#endif
}

bool Nick::hasAll()
{
	return (!host.isNull() && !ident.isNull());
}

Nick::~Nick()
{
}

void recvSetMode(const QString &mode)
{
}

const QString & Nick::getMask()
{
	QString *mask = new QString();
	*mask = QString("%1!*@*").arg(getStringId());
	return *mask;
}
