/*
 * $Id: server.cpp,v 1.111 2004/03/09 13:31:34 blackmor Exp $
 *
 * This file is part of the kwirc project
 * Copyright (C) 2003 nohar, Julien Perrot
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * See the file "COPYING" for the exact licensing terms.
 */
#include <kextsock.h>
#include "config.h"

#include <ktoolbarbutton.h>
#include <kapp.h>
#include <kconfig.h>

#include <qspinbox.h>
#include <qcheckbox.h>
#include <qlayout.h>

#include <stdio.h>
#include <stdlib.h>

#include <kstandarddirs.h>
#include <qpainter.h>
#include <kmenubar.h>
#include <kmessagebox.h>
#include <krootpixmap.h>
#include <kstdaction.h>
#include <klineeditdlg.h>
#include <kdebug.h>
#include <kipc.h>
#include <dcopclient.h>

#include <klocale.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <signal.h>
#include <dirent.h>
#include <fcntl.h>
#include <assert.h>

#include <kiconloader.h>
#include <kstringhandler.h>
#include <ktip.h>

#include "kwircdebug.h"
#include "kwirc.h"
#include "kwircpreferences.h"
#include "server.h"
#include "command.h"
#include "nick.h"
#include "channel.h"
#ifdef FEAT_RUBY
#include "rubyscriptmanager.h"
#endif

#include "kextendedsslsocket.h"
#include <kbufferedio.h>

#include "serversettings.h"
#include "kwircsettings.h"

const Q_ULONG Server::CHAR_BUFFER_SIZE = 128;
const Q_ULONG Server::IRC_MAX_MESSAGE_LENGTH = 512;
// the interval between lag measuring (in seconds)
const Q_ULONG Server::CHECK_INTERVAL = 15;
// maximum lag (in seconds) ; if currentLag > MAX_LAG we assume that we are
// disconnected from the server
const Q_ULONG Server::MAX_LAG = 60;
const QString Server::LAG_TAG = QString("PSYBNCSUCKS");

const Q_ULONG Server::MAX_RETRY = 20;
const Q_ULONG Server::DELAY_BEFORE_RETRY = 10;
const Q_ULONG Server::CONNECT_TIMEOUT = 30;

/*
--------------------------------------------------------------------------------
*/
Server::Server(const QString &host, Q_UINT16 port,
	       const QString &label, const QString &username,
	       const QString & pass,
	       bool ssl, const QString &nick, const QString &configGroup,
	       const QStringList &ignoreList):
	ServerElem((label == QString::null) ? host : label, TYPE_SERVER),
	group(configGroup),
	cur_string(),
	socket(NULL),
	maxModeCount(4),
	chanTypes("#&"),
	nickPrefixes("+@"),
	nicks(28, false),
	yourNick(NULL),
	floodMode(false),
	connecting(false)
{
	retryCounter = 0;
	lookupDone = false;
	charBuffer = (char*)malloc(CHAR_BUFFER_SIZE);

	installEventFilter(this);
	resolver.setRecipient(this);

	floodTimer = new QTimer(this);
	connect(floodTimer, SIGNAL(timeout()), this,
		SLOT(floodTimerTrigger()));

	connect(&notifyTimer, SIGNAL(timeout()),
		this, SLOT (notifyTimeout()) );
	connect(&notifyCheckTimer, SIGNAL(timeout()),
		this, SLOT (notifyCheckTimeout()));
	connect(&connectionTimeoutTimer, SIGNAL(timeout()),
		this, SLOT(handleConnectTimeout()));

	setIgnoreListFromStringList(ignoreList);

#ifdef FEAT_RUBY
	setRubyInstance(RubyScriptManager::getRubyServerInstance(this));
#endif
	connect(this, SIGNAL(logAll(KwircLogType, const QString &,
				const QString &, const QString &)),
			this, SIGNAL(log(KwircLogType, const QString &,
				const QString &, const QString &)));

	changeServer(host, port, username, pass, ssl, nick);
}

Server::~Server()
{
	delete socket;
}

/*******************************
 * raw socket stuff
 * *****************************/

#include <errno.h>

void Server::realPutserv(const QString &str)
{
	if (socket->writeBlock(str.latin1(), str.length()) < 0) {
		kdError(CONNECT_DBG) << "Error in write : " << errno << endl;
		exit(1);
	}
}

void Server::floodTimerTrigger()
{
	sentBytesCount = 0;

	if (floodMode) {
		realPutserv(tbsStrings.first());
		tbsStrings.remove(tbsStrings.first());
		if (tbsStrings.empty())
			floodMode = false;
	}
}

// Basic flood protection
void Server::putservFloodProtection(const QString &str)
{
	sentBytesCount += strlen(str.latin1());

	if (floodMode || sentBytesCount > MAX_SENT_BYTES_COUNT) {
		floodMode = true;
		tbsStrings.append(str);
	} else {
		realPutserv(str);
	}
}

// implmenter un timer
void Server::putserv(const QString &str)
{
	if (str.contains(LAG_TAG)) {
		notifySent.start();
	}
	if (isConnected()) {
		putservFloodProtection(str);
	}
}

void Server::getserv(const QString &str)
{
	if(QString::compare(str.section(' ',0,0),QString("PING")) == 0) {
		QString s = QString("PONG %1\r\n").arg(str.section(' ',1));
		putserv(s);
		return;
	}

	// Our PING's response
	// Does not work with fuckin' psybnc hence the str.contains("PONG")
	//if (str == ":" + host + " PONG " + host + " :" + LAG_TAG) {
	if (QString::compare(str.section(' ',1,1),QString("PONG")) == 0) {
		currentLag = notifySent.elapsed();
		notifyCheckTimer.stop();
		emit lag(currentLag / 1000);
		startNotifyTimer(CHECK_INTERVAL);
		return;
	}
	QString who = str.section(' ',0,0);
	if(who.startsWith(":"))
		who.remove(0, 1);
	QString command = str.section(' ', 1, 1);
	QString what = str.section(' ', 2);
	if(what.startsWith(":"))
		what.remove(0, 1);

	if (str.startsWith("NOTICE AUTH")) {
		doServerNotice(str);
	} else if (QString::compare(command, QString("NOTICE")) == 0) {
		QString s = str.section(' ', 1);
		if (s.startsWith("NOTICE AUTH"))
			doServerNotice(str);
		else
			doNotice(who, what);
	} else if (QString::compare(command, QString("JOIN")) == 0) {
		doJoin(who,what);
	} else if (QString::compare(command, QString("PART")) == 0) {
		doPart(who,what);
	} else if (QString::compare(command, QString("QUIT")) == 0) {
		doQuit(who,what);
	} else if (QString::compare(command, QString("MODE")) == 0) {
		doMode(who,what);
	} else if (QString::compare(command, QString("PRIVMSG")) == 0) {
		doPrivmsg(who,what);
	} else if (QString::compare(command, QString("NICK")) == 0) {
		doNick(who,what);
	} else if (QString::compare(command, QString("INVITE")) == 0) {
		doInvite(who,what);
	} else if (QString::compare(command, QString("KICK")) == 0) {
		doKick(who,what);
	} else if (QString::compare(command, QString("KILL")) == 0) {
		doKill(who,what);
	} else if (QString::compare(command, QString("TOPIC")) == 0) {
		doTopic(who,what);
	} else if (QString::compare(command, QString("433")) == 0) {
		do433(who,what);
	} else if (QString::compare(command, QString("001")) == 0) {
		do001(who,what);
	} else if (QString::compare(command, QString("005")) == 0) {
		do005(who,what);
	} else if (QString::compare(command, QString("301")) == 0) {
		do301(who,what);
	} else if (QString::compare(command, QString("307")) == 0) {
		do307(who,what);
	} else if (QString::compare(command, QString("401")) == 0) {
		do401(who,what);
	} else if (QString::compare(command, QString("305")) == 0) {
		do305(who,what);
	} else if (QString::compare(command, QString("306")) == 0) {
		do306(who,what);
	} else if (QString::compare(command, QString("315")) == 0) {
		do315(who,what);
	} else if (QString::compare(command, QString("332")) == 0) {
		do332(who,what);
	} else if (QString::compare(command, QString("333")) == 0) {
		do333(who,what);
	} else if (QString::compare(command, QString("352")) == 0) {
		do352(who,what);
	} else if (QString::compare(command, QString("353")) == 0) {
		do353(who,what);
	} else if (QString::compare(command, QString("366")) == 0) {
		do366(who,what);
	} else if (QString::compare(command, QString("376")) == 0) {
		do376();
	} else if (QString::compare(command, QString("311")) == 0) {
		do311(who,what);
	} else if (QString::compare(command, QString("312")) == 0) {
		do312(who,what);
	} else if (QString::compare(command, QString("313")) == 0) {
		do313(who,what);
	} else if (QString::compare(command, QString("317")) == 0) {
		do317(who,what);
	} else if (QString::compare(command, QString("318")) == 0) {
		do318(who,what);
	} else if (QString::compare(command, QString("319")) == 0) {
		do319(who,what);
	} else if (QString::compare(command, QString("275")) == 0) {
		do275(who,what);
	} else if (QString::compare(command, QString("324")) == 0) {
		do324(who,what);
	} else if (QString::compare(command, QString("367")) == 0) {
		do367(who,what);
	} else if (QString::compare(command, QString("368")) == 0) {
		do368(who,what);
	} else if (QString::compare(command, QString("ERROR")) == 0) {
		doError(who,what);
	} else {
		doElse(who,command,what);
	}
}

void Server::do001(QString &who, QString &what)
{
	yourNick = nickFromString(what.section(' ', 0, 0 ));
	serverString = who;
	emit log(LOG_RAW, getStringId(), what, QString::null);
	// XXX
	if (!autojoin.isEmpty())
		sendJoin(autojoin, autojoinkeys);

	startNotifyTimer(CHECK_INTERVAL);
}

void Server::do005(QString &, QString &what)
{
	QRegExp re;
	re.setPattern("MODE=(\\d+)");
	if (re.search(what) != -1)
		maxModeCount = re.cap(1).toInt();
	re.setPattern("CHANTYPES=(\\S+)");
	if (re.search(what) != -1)
		chanTypes = re.cap(1);
	re.setPattern("PREFIX=(\\S+)");
	if (re.search(what) != -1)
		nickPrefixes = re.cap(1);
}

void Server::do275(QString &w, QString &what)
{
	doWhois(w, what);
}

void Server::do311(QString &w, QString &what)
{
	doWhois(w, what);
}
void Server::do312(QString &w, QString &what)
{
	doWhois(w, what);
}
void Server::do313(QString &w, QString &what)
{
	doWhois(w, what);
}
void Server::do301(QString &w, QString &what)
{
	doWhois(w, what);
}
void Server::do307(QString &w, QString &what)
{
	doWhois(w, what);
}
void Server::do401(QString &w, QString &what)
{
	doWhois(w, what);
}

//kwirc (in): :irc.iiens.net 305 nohar` :You are no longer marked as being away
void Server::do305(QString &, QString &)
{
	logActive(LOG_RAW, getStringId(),
			i18n("You are no longer marked as being away"),
			QString::null);
}
//kwirc (in): :irc.iiens.net 306 nohar` :You have been marked as being away
void Server::do306(QString &, QString &)
{
	logActive(LOG_RAW, getStringId(),
			i18n("You have been marked as being away"),
			QString::null);
}
void Server::do315(QString &, QString &what)
{
	QString chan = what.section(' ', 1, 1);
	Channel *c = channel(chan, false);
	if (c) {
		c->setWhoDisplay(true);
		c->endOfWho();
	}
}

void Server::do317(QString &, QString &what)
{
	QString seconds = what.section(' ', 2, 2);
	unsigned long long nsecs = seconds.toULong();
	unsigned int days = nsecs / (unsigned long long)(60*60*24);
	nsecs -= days * 60 * 60 * 24;
	unsigned int hours = nsecs / (60 * 60);
	nsecs -= hours * (60 * 60);
	unsigned int minutes = nsecs / 60;
	nsecs -= minutes * 60;

	QString info = QString(i18n("Idle since: %1d %2h %3m %4s"))
		.arg(days).arg(hours)
		.arg(minutes).arg((unsigned int)nsecs);
	emit logActive(LOG_WHOIS, getStringId(), info, QString::null);
}

void Server::do318(QString &w, QString &what)
{
	doWhois(w, what);
}
void Server::do319(QString &w, QString &what)
{
	doWhois(w, what);
}


//kwirc (in): :irc.iiens.net 324 coq #kwirc +tnr
void Server::do324(QString &, QString &what)
{
	QString chan = what.section(' ', 1, 1);
	QString modes = what.section(' ', 2);
	Channel *c = channel(chan);

	if(!c)
		return;
	c->setInitialModes(modes);
}


/*
 *         332     RPL_TOPIC
                        "<channel> :<topic>"

                - When sending a TOPIC message to determine the
                  channel topic, one of two replies is sent.  If
                  the topic is set, RPL_TOPIC is sent back else
                  RPL_NOTOPIC.
 */
void Server::do332(QString &who, QString &what)
{
	QString chan = what.section(' ',1,1);
	QString topic = what.section(' ',2);
	if(topic.startsWith(":"))
		topic.remove(0,1);
	Channel *c = channel(chan);
	if(!c)
		return;
	c->recvTopic(who, topic);
}

void Server::do333(QString &, QString &what)
{
	QString chan = what.section(' ',1,1);
	QString whoset = what.section(' ',2,2);
	QString when = what.section(' ',3,3);
	Channel *c = channel(chan);
	if(!c)
		return;
	c->setWhoSet(whoset, when);
}

// :irc.iiens.net 352 Kwirc` #insomnia ~nohar-iie m235.net81-67-209.noos.fr irc.iiens.net nohar`emacs`power`user HS@ :0 nohar

// FIXME: wont work

void Server::do352(QString &, QString &what)
{
	QString whoRepl = what.section(' ', 1);
	QString str = what.section(' ', 1, 1);
	QString ident = what.section(' ', 2, 2);
	QString host = what.section(' ', 3, 3);
	QString server = what.section(' ', 4, 4);
	QString nick = what.section(' ', 5, 5);
	QString auxInfos = what.section(' ', 6, 6);

	Nick *n = nickFromString(nick);
	if (!n->hasAll()) {
		n->setHost(host);
		n->setIdent(host);
	}
	if (auxInfos.contains('H'))
		n->setAway(false);
	if (auxInfos.contains('G'))
		n->setAway(true);

	if (Channel *c = channel(str, false)) {
		c->recvWho(n, whoRepl);
	} else
		logActive(LOG_WHO, getStringId(), nick, whoRepl);
}

void Server::do353(QString &, QString &what)
{
	//TODO: passer en argument a channel la chaine a parser.
	QString chan = what.section(' ', 2,2);
	Channel *c;
	c = channel(chan);
	if(!c)
		return;

	QString nicklist = what.section(' ',3);
	if(nicklist.startsWith(":"))
		nicklist.remove(0,1);
	c->setInChannel(nicklist);
}

void Server::do366(QString &, QString &what)
{
	QString chanName = what.section(' ',1,1);
	Channel *c = channel(chanName);
	if(!c)
		return;
	c->sync();
}

// kwirc (in): :irc.iiens.net 367 coq #kwirc *!*@*.xchat.users.org
// Blackmore!jipe@81.56.234.233 1065114525
void Server::do367(QString &,QString &what)
{
	QString chan = what.section(' ', 1, 1);
	QString mask = what.section(' ', 2, 2);
	Channel *c = channel(chan);

	if(!c)
		return;

	c->addInitialBanMask(mask);
}

//kwirc (in): :irc.iiens.net 368 coq #kwirc :End of Channel Ban List
void Server::do368(QString &,QString &)
{

}

/* end of motd */
void Server::do376()
{
	connecting = false;
}

// XXX: fixme
void Server::do433(QString &, QString &)
{
	if (connecting) {
		QString arg(i18n("your nick is already taken! i add a `"));
		emit log(LOG_RAW, getStringId(), arg, QString::null);
		nick_tosend.append("`");
		sendNick(nick_tosend);
		if(yourNick)
			removeNick(yourNick);
		yourNick = nickFromString(nick_tosend);
	} else {
		emit logAll(LOG_RAW, getStringId(), "This nick is already"
				" taken, choose another one", "");
	}
}

void Server::doElse(QString &, QString &, QString &what)
{
	if(!what.isNull())
		emit log(LOG_RAW, getStringId(), what, QString::null);
}

void Server::doError(QString &who, QString &what)
{
	emit log(LOG_RAW, getStringId(),
			QString("ERROR:").append(who).append(what),
			QString::null);
	socket->close();
	delete socket;
	socket = NULL;
}

void Server::doInvite(QString &who, QString &what)
{
	QString invited = what.section(' ', 0,0);
	QString chan = what.section(' ', 1,1);

	Nick *n = nickFromString(who);
	if (isIgnored(n->getIrcId(), false, false, false, true))
	    return;

	Nick *inv = nickFromString(invited);

	if (KwircSettings::joinOnInvite()
			&& inv == getYourNick()) {
		sendJoin(chan);
		emit logActive(LOG_INVITE,
				getStringId(), QString("you"), chan);
	} else {
		Channel *c = channel(chan);
		if(!c)
			return;

		n->recvInvite(inv, c);
	}
}

void Server::doJoin(QString &who, QString &what)
{
	if (what.length() < 1
			|| (!what.startsWith("#") && !what.startsWith("&")))
		return;

	Nick *n = nickFromString(who);
	Channel *c = channel(what, false);

	if(n == getYourNick()) {
		// we were already in this channel, happens with psybnc on
		// reconnection. Cleanup
		// XXX: we should do this evervy time (even if c is new) do a
		// /names or /who automatically
		if (!c)
			c = channel(what.section(' ', 0, 0), true);
		if (!c)
			return;
		c->youJoined();
	} else if (c) {
		n->recvJoin(c);
	}
}

void Server::doKick(QString &who, QString &what)
{
	QString chan = what.section(' ', 0, 0);
	if(chan.length() < 1 && !(chan.startsWith("#")||chan.startsWith("&")))
		return;
	what = what.section(' ', 1);
	if(what.startsWith(":"))
		what.remove(0, 1);

	QString kicked = what.section(' ', 0, 0);
	what = what.section(' ', 1);
	if(what.startsWith(":"))
		what.remove(0, 1);

	Channel *c = channel(chan);
	if(!c)
		return;
	Nick *nkicked = nickFromString(kicked);
	Nick *kicker = nickFromString(who);
	if(nkicked == getYourNick()) {
		c->youWereKicked(kicker, what);
		removeChannel(c);
		delete c;
	} else {
		nkicked->recvKick(kicker, c, what);
		if (!nkicked->used()) {
			removeNick(nkicked);
			nkicked->toBeDeleted();
			delete nkicked;
		}
	}
}

void Server::deleteNick(Nick *n)
{
	removeNick(n);
	n->toBeDeleted();
	delete n;
}

void Server::doKill(QString &who, QString &what)
{
	QString killed = what.section(' ',0,0);
	what = what.section(' ',1);
	if(what.startsWith(":"))
		what.remove(0,1);
	Nick *n = nickFromString(who);
	n->recvKill(what);
	removeNick(n);
	n->toBeDeleted();
	delete n;
}

void Server::doMode(QString &who, QString &what)
{
	QString moded = what.section(' ', 0, 0);
	QString mode = what.section(' ', 1);

	if(moded.startsWith("#") || moded.startsWith("&")) {
		Nick *nick = nickFromString(who);
		Channel *c = channel(moded);
		if(!c)
			return;
		c->recvMode(nick, mode);
	} else {
		Nick *nick = nickFromString(moded);
		emit log(LOG_MODE, getStringId(), nick->getName(), moded);
	}
}

void Server::doNick(QString &who, QString &what)
{
	if (!getYourNick()) {
		yourNick = nickFromString(what.section(' ', 0, 0));
		return;
	}
	if (what == getYourNick()->getStringId()) // psybnc hack
		return;

	Nick *n = nickFromString(who);
	removeNick(n);
	n->recvNick(what);
	insertNick(n);
}

void Server::doServerNotice(const QString &str)
{
	emit log(LOG_NOTICE, getStringId(), str, "");
}

void Server::doNotice(QString &who, QString &what)
{
	QString message = what.section(' ', 1);
	QString to = what.section(' ', 0, 0);
	if(message.startsWith(":"))
		message.remove(0, 1);
	Nick *nick = nickFromString(who);
	if (isIgnored(nick->getIrcId(), false, false, true, false))
		return;
	if(to.length() < 1)
		return;

	if(to.startsWith("#") || to.startsWith("&")) {
		doPublicNotice(nick, channel(to), message);
	} else if(nickFromString(to) == getYourNick()) {
		doPrivateNotice(nick, message);
	}
}

void Server::doPart(QString &who, QString &what)
{
	QString chan = what.section(' ', 0, 0);
	what = what.section(' ', 1);
	if(what.startsWith(":"))
		what.remove(0,1);	// get rid of ':'
	if(what.isNull())
		what = "";
	Nick *n = nickFromString(who);
	Channel *c = channel(chan);
	if(!c)
		return;

	n->recvPart(c, what);

	if(n == getYourNick()) {
		c->youLeft(what);
		removeChannel(c);
		c->toBeDeleted();
		delete c;
	} else {
		if (!n->used()) {
			removeNick(n);
			n->toBeDeleted();
			delete n;
		}
	}
}

void Server::doPrivate(Nick *nick, QString &message)
{
	if (ctcpMessage(message)) {
		// X-QUOTE stuff CTCP stuff goes here
		// replies are notice commands..
		QString ctcp = message.section(' ', 0, 0);
		message = message.section(' ', 1);
		if (ctcp.compare("ACTION") == 0) {
			emit newQuery(nick);
			nick->recvAction(message);
		} else {
			nick->recvCtcp(ctcp, message);
		}
	} else {
		emit newQuery(nick);
		nick->recvPrivmsg(message);
	}
}

void Server::doPrivateNotice(Nick *nick, QString &message)
{
	if(ctcpMessage(message)) {
		// X-QUOTE stuff CTCP replies go here
		// replies are notice commands..
		QString msg = QString("CTCP reply: %1 %2")
			.arg(nick->getStringId()).arg(message);
		emit logActive(LOG_CTCP, nick->getStringId(),
				msg, QString::null);
	} else
		nick->recvNotice(message);
}

void Server::doPrivmsg(QString &who, QString &what)
{
	QString message = what.section(' ', 1);
	QString to = what.section(' ', 0, 0);
	if(message.startsWith(":"))
		message.remove(0,1);
	Nick *nick = nickFromString(who);
	QString ircId = nick->getIrcId();

	if(to.length() < 1)
		return;

	//*** normal messages ****
	if(to.startsWith("#") || to.startsWith("&")) {
		if (!isIgnored(ircId, true, false, false, false))
			doPublic(nick, channel(to), message);
	} else if (nickFromString(to) == getYourNick()) {
		if (!isIgnored(ircId, false, true, false, false))
			doPrivate(nick, message);
	} else if (nick == getYourNick()) {
		// proxy tells us other client sends private message
		// XXX: CTRLPROXY SUCKS
		QString msg = QString("<%1> %2").arg(nick->getStringId())
			.arg(message);
		doPrivate(nickFromString(to), msg);
	}
}

void Server::doPublic(Nick *nick, Channel *channel, QString &message)
{
	if (!nick || !channel) {
		log(LOG_RAW, getStringId(),
				"is this an irc server?", QString::null);
		return;
	}
	if(ctcpMessage(message)) {
		//***********************************
		// X-QUOTE stuff CTCP stuff goes here
		//***********************************
		// replies are notice commands..
		QString ctcp = message.section(' ',0,0);
		message = message.section(' ',1);
		if(ctcp.compare("ACTION") == 0)
			channel->recvAction(nick, message);
		else
			channel->recvCtcp(nick, ctcp, message);
	} else
		channel->recvPrivmsg(nick, message);
}


void Server::doPublicNotice(Nick *nick, Channel *chan, QString &message)
{
	if(ctcpMessage(message)) {
		// X-QUOTE stuff CTCP replies go here
		// replies are notice commands..
		QString ctcp = message.section(' ', 0, 0);
		message = message.section(' ', 1);
		chan->recvCtcpNotice(nick, ctcp, message);
	} else
		chan->recvNotice(nick, message);
}

//:tpvmc!~test1@laoshi.estu.iie.cnam.fr QUIT :Client closed connection
void Server::doQuit(QString &who, QString &what)
{
	Nick *n = nickFromString(who);
	n->recvQuit(what);
	removeNick(n);
//:nohar!~nohar@nucleus.estu.iie.cnam.fr PRIVMSG #37337 :yop
	n->toBeDeleted();
	delete n;
}

void Server::doTopic(QString &who, QString &what)
{
	QString chan = what.section(' ', 0,0);
	if(chan.startsWith(":"))
		chan.remove(0,1);
	QString topic = what.section(' ', 1);
	if(topic.startsWith(":"))
		topic.remove(0,1);
	Channel *c = channel(chan);
	if(!c)
		return;
	c->recvTopic(who, topic);
//	nickFromString(who)->recvTopic(c, topic);
}

void Server::doWhois(QString &, QString &what)
{
	QString info = what.section(" ",1);
	emit logActive(LOG_WHOIS, getStringId(), info, QString::null);
}

bool Server::ctcpMessage(QString &message)
{
	if (message.startsWith("\001")) {
		message.remove(message.length() - 1, 1);	// last \001
		message.remove(0, 1);				// first \001
		return true;
	}
	return false;
}

/*****************************************
 * Helper funcs, ui interaction
 *****************************************/

// kwirc!~test1@laoshi.estu.iie.cnam.fr
// a nick string can be simply a nick or a bli!blah@blouh
Nick *Server::nickFromString(const QString &astr, bool create)
{
	QString str = astr;
	QString nickStr;
	bool hasInfo = false;
	Nick *n = NULL;
	while(str.startsWith("@") || str.startsWith("+"))
		str.remove(0,1);

	if(str.contains('!')) {
		nickStr = str.section('!', 0, 0);
		hasInfo = true;
	} else {
		nickStr = str;
	}

	n = nicks[nickStr];
	if (!n && !create)
		return NULL;
	if(!n) {
		n = new Nick(this, str);
		nicks.insert(nickStr, n);
		connect(this, SIGNAL(logAll(KwircLogType, const QString &,
					const QString &, const QString &)),
				n, SIGNAL(log(KwircLogType, const QString &,
						const QString &,
						const QString &)));
	} else if (hasInfo && !n->hasAll()) {
		n->setInfoString(str);
	}
	return n;
}

void Server::removeNick(Nick *n)
{
	QDictIterator<Nick> it(nicks); // See QDictIterator
	nicks.remove(n->getName());
}

void Server::insertNick(Nick *n)
{
	nicks.insert(n->getName(), n);
}

void Server::removeChannel(Channel *c)
{
	QString key = c->getStringId();
	key = key.lower();
	channels.remove(key);
}

Channel *Server::channel(const QString &what, bool create)
{
	Channel *c;
	QString key = what.lower();
	if((c = channels[key]))
		return c;
	if(create) {
		if (what.length() > 200
			&& (!what.startsWith("&") || !what.startsWith("#"))
			&& what.contains(" ,") && what.contains(QChar(7)))
			return NULL;
		c = new Channel(this, what);
		channels.insert(key, c);
		emit newChannel(c);
		connect(this, SIGNAL(logAll(KwircLogType, const QString &,
					const QString &, const QString &)),
				c, SIGNAL(log(KwircLogType, const QString &,
						const QString &,
						const QString &)));
	}
	return c;
}

/** Sends **/

void Server::sendWho(const QString &a)
{
	putserv(QString("WHO %1\r\n").arg(a));
}

void Server::sendAway(const QString &a)
{
	putserv(QString("AWAY :%1\r\n").arg(a));
}

void Server::sendWhois(const QString &a)
{
	putserv(QString("WHOIS %1\r\n").arg(a));
}

class Mode
{
	private:
		QChar c;
		QString arg;
	public:
		Mode(const QChar &, const QString &);
		const QChar &getChar();
		const QString &getArg();
};

Mode::Mode(const QChar &ac, const QString &as)
	: c(ac),
	arg(as)
{
}

const QChar &Mode::getChar()
{
	return c;
}

const QString &Mode::getArg()
{
	return arg;
}

class Modes
{
	private:
		QPtrList<Mode> add;
		QPtrList<Mode> rem;
	public:
		Modes(const QString &modes, const QString &args);
		QString renderMode(int, QString &m, QString &a);
};

Modes::Modes(const QString &modeChars, const QString &args)
{
	QStringList argList = QStringList::split(' ', args);
	QStringList::Iterator currentArg = argList.begin();
	bool adding = true;

	for(unsigned int i = 0; i < modeChars.length(); i++) {
		switch(modeChars.at(i).latin1()) {
		case '+':
			adding = true;
			break;
		case '-':
			adding = false;
			break;
		case 'b':
		case 'o':
		case 'v':
			if(adding)
				add.append(new Mode(modeChars.at(i),
						    *currentArg));
			else
				rem.append(new Mode(modeChars.at(i),
							    *currentArg));
			currentArg++;
			break;
		case 'p':
		case 's':
		case 'i':
		case 't':
		case 'n':
		case 'm':
			if(adding)
				add.append(new Mode(modeChars.at(i),
						    QString::null));
			else
				rem.append(new Mode(modeChars.at(i),
						    QString::null));
			currentArg++;
			break;
		case 'l':
		case 'k':
			if(adding)
				add.append(new Mode(modeChars.at(i),
						    *currentArg));
			else
				rem.append(new Mode(modeChars.at(i),
						    *currentArg));
			currentArg++;
			break;
		default:
			break;
		}
	}
}

QString Modes::renderMode(int num, QString &mode, QString &args)
{
	bool modeRemoving = false;
	for (int i = 0; i < num ; i++) {
		if (add.count() > 0) {
			if (i == 0)
				mode += '+';
			Mode *m = add.at(0);
			add.remove(m);
			mode += m->getChar();
			if (!m->getArg().isEmpty() && !args.isEmpty())
				args += ' ';
			args += m->getArg();
			delete m;
		}
		if (rem.count() > 0) {
			if (i == 0 || !modeRemoving) {
				mode += '-';
				modeRemoving = true;
			}
			Mode *m = rem.at(0);
			rem.remove(m);
			mode += m->getChar();
			if (!m->getArg().isEmpty() && !args.isEmpty())
				args += ' ';
			args += m->getArg();
			delete m;
		}
	}
	if (mode.isNull())
		return QString::null;
	if (args.isNull())
		return mode;
	return mode + " " + args;
}

void Server::sendMode(const QString &chan, const QString &mode,
		const QString &args)
{
//(:nohar!~nohar@nucleus.estu.iie.cnam.fr MODE #37337 -i )
	Modes modes(mode, args);
	QString m = QString::null;
	QString a = QString::null;
	modes.renderMode(maxModeCount, m, a);
	while(m != QString::null) {
		putserv(QString("MODE %1 %2 %3\r\n").arg(chan).arg(m).arg(a));
		m = QString::null;
		a = QString::null;
		modes.renderMode(maxModeCount, m, a);
	}
}

void Server::sendCtcpNotice(const QString &dest, const QString &what)
{
	putserv(QString("NOTICE %1 :\001%2\001\r\n").arg(dest).arg(what));
}

void Server::sendCtcp(const QString &dest, const QString &what)
{
	putserv(QString("PRIVMSG %1 :\001%2\001\r\n").arg(dest).arg(what));
}

void Server::sendJoin(const QString &strs, const QString &keys)
{
	if(keys.isNull())
		putserv(QString("JOIN :%1\r\n").arg(strs));
	else
		putserv(QString("JOIN %1 :%2\r\n").arg(strs).arg(keys));
}

void Server::sendTopic(const QString &channel, const QString &topic)
{
	putserv(QString("TOPIC %1 :%2\r\n").arg(channel).arg(topic));
}

void Server::sendNick(const QString &nick)
{
	putserv(QString("NICK :%1\r\n").arg(nick));
}

void Server::sendPart(const QString &chan, const QString &msg)
{
	QString partMessage = msg;
	if (partMessage.isEmpty())
		partMessage = KwircSettings::partMessage();
	putserv(QString("PART %1 :%2\r\n").arg(chan).arg(partMessage));
}

void Server::sendUser(const QString &username,
		const QString &hostname,
		const QString &servername,
		const QString &realname)
{
	QString tosend = QString("USER %1 %2 %3 :%4\r\n").arg(username)
		.arg(hostname).arg(servername).arg(realname);
	putserv(tosend);
}

void Server::sendOper(const QString &username, const QString &hostname)
{
	putserv(QString("OPER %1 :%2\r\n").arg(username).arg(hostname));
}

void Server::sendPass(const QString &p)
{
	putserv(QString("PASS :%1\r\n").arg(p));
}

void Server::sendPrivmsg(const QString &to,const QString &msg)
{
	putserv(QString("PRIVMSG %1 :%2\r\n").arg(to).arg(msg));
}

void Server::sendAction(const QString &to,const QString &msg)
{
	sendCtcp(to, QString("ACTION ") + msg);
}

void Server::sendQuit(const QString &msg)
{
	if (socket) {
		QString quitMessage = msg;
		if (quitMessage.isNull())
			quitMessage = KwircSettings::quitMessage();
		putserv(QString("QUIT :%1\r\n").arg(quitMessage));
		emit logAll(LOG_RAW, getStringId(), "Disconnected. You quit: ",
			quitMessage);
		// XXX we should wait for the socket to be closed
		cleanup();
	}
}

/****************************************
 * connection handling
 ***************************************/

bool Server::changeServer(const QString &ahost, Q_UINT16 aport,
		const QString &username, const QString &pass, bool assl,
		const QString &nick)
{
	if (isConnected()) {
		sendQuit(i18n("Changing servers"));
	}
	if (nick.isEmpty()) {
		nick_tosend = KwircSettings::nickName();
	} else {
		nick_tosend = nick;
	}
	if (username.isEmpty()) {
		username_tosend = KwircSettings::userName();
	} else {
		username_tosend = username;
	}
	serverString = host;
	pass_tosend = pass;

	if (socket) {
		socket->close();
		delete socket;
		socket = NULL;
	}
	host = ahost;
	port = aport;
	ssl = assl;

	createSocket();
	return true;
}

void Server::makeConnect()
{
	logInfoMsg(i18n("Connection to %1 successful").arg(host));
	// Connection is successfull, reseting the retry counter
	retryCounter = 0;
	socket->enableRead(true);
	floodTimer->start(2000, false);

	if(!pass_tosend.isEmpty())
		sendPass(pass_tosend);
	sendNick(nick_tosend);
	//yourNick = nickFromString(nick_tosend);
	sendUser(username_tosend, KwircPreferences::getHostname(),
			getStringId(), KwircSettings::realName());
}

void Server::handleConnectionFailure(int error)
{
	kdError(SERVER_DBG) << "Connection failure" << endl;
	logErrorMsg(i18n("Connection to %1  failed (Error number:%2)")
			.arg(host).arg(error));

	retryCounter++;
	if (socket) {
		delete socket;
		socket = NULL;
	}
	doReconnect();
}

void Server::socketClosed(int state)
{
	if (socket && state == KBufferedIO::involuntary) {
		kdWarning(CONNECT_DBG) << "disconnected" << endl;
		logErrorMsg(i18n("Connection to %1 was closed").arg(host));
		retryCounter++;
		doReconnect();
	}
}

void Server::serverCloses()
{
// FIXME delete stuff
}

bool Server::eventFilter(QObject* parent,QEvent* event)
{
	if(event->type()==QEvent::User)
	{
		lookupFinished();
		return true;
	}
	return ServerElem::eventFilter(parent, event);
}

void Server::createSocket()
{
	int flags = KExtendedSocket::inetSocket
		    | KExtendedSocket::inputBufferedSocket;

#ifdef HAVE_SSL
	if (ssl) {
		socket = new KExtendedSSLSocket(host, port, flags );
	} else
#endif
	socket = new KExtendedSocket(host, port, flags);

	resolver.setSocket(socket);
	resolver.setRecipient(this);

	connect(socket, SIGNAL(readyRead()), SLOT(serverTalks()) );
	connect(socket, SIGNAL(connectionSuccess()),
			SLOT(handleConnectionSuccess()) );
	connect(socket, SIGNAL(closed(int)), SLOT(socketClosed(int)));
	connect(socket, SIGNAL(connectionFailed(int)),
			SLOT(handleConnectionFailure(int)));
#ifdef HAVE_SSL
	if (ssl) {
		connect(socket, SIGNAL(ssl_handshaken(bool)),
			SLOT(handleSSLHandshake(bool)));
	}
#endif
}

void Server::cleanup()
{
	if (yourNick != NULL)
		nick_tosend = yourNick->getStringId();

	notifyCheckTimer.stop();
	if (socket) {
		socket->close();
		delete socket;
		socket = NULL;
	}

	floodTimer->stop();
	floodMode = false;
	notifyCheckTimer.stop();
	notifyTimer.stop();
	tbsStrings.clear();

	/*channels must be first since they abuse of their nicks in their
	 * deletor */
	autojoin = "";
	QDictIterator<Channel> itt( channels ); // See QDictIterator
	for( ; itt.current(); ++itt ) {
		autojoin += (autojoin.isEmpty()?"":",")
			+ itt.current()->getStringId();
		Channel *c =  itt.current();
		c->toBeDeleted();
		delete c;
	}
	channels.clear();

	QDictIterator<Nick> it(nicks); // See QDictIterator
	for( ; it.current(); ++it ) {
		Nick *nick = it.current();
		if (!nick) {
			kdError() << k_funcinfo << "NULL nick : " <<
				it.currentKey() << endl;
			continue;
		}
		nick->toBeDeleted();
		delete nick;
	}
	nicks.clear();

	yourNick = NULL;
}

void Server::doReconnect()
{
	if (socket) {
		emit logAll(LOG_RAW, getStringId(), "Disconnected. Connection"
				" reset by peer", "");
		cleanup();
	}

	createSocket();

	logInfoMsg(i18n("Reconnecting in %1 seconds").arg(DELAY_BEFORE_RETRY));
	QTimer::singleShot(1000 * DELAY_BEFORE_RETRY, this,
			SLOT(startConnect()));
}

// Timer callbacks used to calculate self-lag
void Server::notifyTimeout()
{
	if (notifyCheckTimer.isActive())
		return;

	putserv(QString("PING " + LAG_TAG + "\r\n"));
	notifyCheckTimer.start(MAX_LAG * 1000);
}

void Server::notifyCheckTimeout()
{
	currentLag = MAX_LAG * 1000; // for consistency
	emit lag(currentLag / 1000);
	logErrorMsg(i18n("Connection to %1 lost").arg(host));
	doReconnect();
}

void Server::startNotifyTimer(int sec)
{
	// make sure the timer gets started properly in case we have
	// reconnected
	notifyTimer.stop();
	// start the timer in one shot mode
	notifyTimer.start(sec * 1000, true);
}



/* *****************
 * otherstuff
 * ******************/

void Server::serverTalks()
{
	Q_LONG sz;
	unsigned long bytesRead = 0;
	unsigned long numLines = 0;

	while ((sz = socket->readBlock(charBuffer, CHAR_BUFFER_SIZE)) >= 0) {
		Q_ULONG size = (Q_ULONG)sz;
		if (size == 0)
			return;
		bytesRead += size;
		charBuffer[size] = 0;
		if (strlen(charBuffer) < size) {
			kdError(SERVER_DBG) << "zero inside string " <<
				charBuffer << endl;
			exit(1);
		}

		QString instr(cur_string);	// get last unprocessed string
		cur_string = QString::null;

		instr.append(QString(charBuffer));
		if(! instr.endsWith("\n")) {
			cur_string = instr.section("\n", -1);
			instr.setLength(instr.length() - cur_string.length());
		}
		QStringList ql = QStringList::split("\n", instr);
		QStringList::iterator it;
		numLines += ql.size();
		for ( it = ql.begin(); it != ql.end(); ++it ) {
			if((*it).endsWith("\r"))
				(*it).setLength((*it).length() - 1);
			getserv(*it);
		}
		/*
		if (bytesRead >= IRC_MAX_MESSAGE_LENGTH * 3 || numLines > 10) {
			idler->start(100, TRUE);
			return;
		}*/
	}
}

Nick *Server::getYourNick()
{
	if(yourNick == NULL)
		kdError(SERVER_DBG) << "error : yourNick" << endl;
	return yourNick;
}

void Server::slotQuery(const QString &nick)
{
	emit newQuery(nickFromString(nick));
}

void Server::setAutoJoin(QString aj, QString ajk)
{
	autojoin = aj;
	autojoinkeys = ajk;
}

bool Server::isConnected()
{
	if ( socket == NULL ) {
		kdError(SERVER_DBG) << "error : socket is null" << endl;
		return false;
	}
	return socket->socketStatus() == KExtendedSocket::connected;
}


void Server::logErrorMsg(const QString &msg)
{
	logMsg(LOG_ERROR, msg);
}

void Server::logInfoMsg(const QString &msg)
{
	logMsg(LOG_INFO, msg);
}

void Server::logMsg(KwircLogType level, const QString &msg)
{
	emit log(level,  getStringId(),  msg,  QString::null);
}

IgnoreListItem::IgnoreListItem(bool achan, bool apriv, bool anotice,
		bool ainvite) :
	chan(achan),priv(apriv),notice(anotice),invite(ainvite)
{}

QString IgnoreListItem::toPrettyString()
{
	QString res = QString("");
	if (chan)
		res.append("CHAN ");
	if (priv)
		res.append("PRIV ");
	if (notice)
		res.append("NOTICE ");
	if (invite)
		res.append("INVITE");

	return res;
}

QString IgnoreListItem::toString()
{
	QString res = QString();
	bool b[4];
	b[0] = chan; b[1] = priv; b[2] = notice; b[3] = invite;

	for (int i=0; i<4; i++) {
		char c = b[i] ? '1' : '0';
		res.append(c);
	}

	return res;
}

IgnoreListItem IgnoreListItem::itemFromString(const QString &s, bool *r)
{
	IgnoreListItem item;
	bool b[4];

	for (int i=0; i<4; i++) {
		if (s[i] == '1') {
			b[i] = true;
		} else if (s[i] == '0') {
			b[i] = false;
		} else {
			*r = false;
			return item;
		}
	}

	*r = true;
	item.chan = b[0]; item.priv = b[1];
	item.notice = b[2]; item.invite = b[3];

	return item;
}

void Server::addIgnoreMask(const QString &mask, bool chan, bool priv,
		bool notice, bool invite)
{
	bool isIrcId = false;
	QString realMask;
	Nick *n = nicks[mask];

	isIrcId = mask.contains('*')
		|| mask.contains('@') || mask.contains('!');

	if (isIrcId) {
		realMask = mask;
	} else if (n != NULL) {
		realMask = n->getMask();
	} else {
		kdWarning(SERVER_DBG) << mask << " is not a ircid mask nor "
			<< "an existing nick" << endl;
		return;
	}

	if (!ignoreList.contains(realMask)) {
		ignoreList[realMask]= IgnoreListItem(chan, priv, notice,
				invite);
	} else {
		logInfoMsg(i18n("Mask %1 already in ignore list")
				.arg(realMask));
	}
}

void Server::setIgnoreMask(const QString &mask, bool chan, bool priv,
		bool notice, bool invite)
{
	if (ignoreList.contains(mask)) {
		ignoreList[mask].chan   = chan;
		ignoreList[mask].priv   = priv;
		ignoreList[mask].notice = notice;
		ignoreList[mask].invite = invite;
	}
}

void Server::removeIgnoreMask(const QString &mask)
{
	if (ignoreList.contains(mask)) {
		ignoreList.remove(mask);
	} else {
		logInfoMsg(i18n("Ignore mask %1 not found").arg(mask));
	}
}

bool Server::isIgnored(const QString &ircid, bool chan, bool priv, bool notice,
		bool invite)
{
	IgnoreList::Iterator it;
	for(it = ignoreList.begin() ; it != ignoreList.end(); ++it ) {
		IgnoreListItem item = it.data();
		QString mask = it.key();

		if ((chan && !item.chan)
		    || (priv && !item.priv)
		    || (notice && !item.notice)
		    || (invite &&!item.invite))
			continue;

		// We want a regexp case-insensitive and
		// in wildcard mode
		QRegExp re = QRegExp(mask, false, true);
		if (re.exactMatch(ircid)) {
			return true;
		}
	}
	return false;
}

const IgnoreList &Server::getIgnoreList()
{
	return ignoreList;
}

const QDict<Channel> &Server::getChannels()
{
	return channels;
}

const QDict<Nick> &Server::getNicks()
{
	return nicks;
}

void Server::setIgnoreListFromStringList(QStringList l)
{
	ignoreList.clear();
	for (QStringList::Iterator it = l.begin(); it != l.end(); ++it ) {
		QString s = *it;
		bool res = false;
		IgnoreListItem item;
		uint length = s.length();
		QString mask = s.left(length-4);
		if (mask.isEmpty()) {
			kdWarning() << k_funcinfo << "Mask is empty for " << s << endl;
			continue;
		}
		item = IgnoreListItem::itemFromString(s.right(4), &res);
		if (!res) {
			kdWarning() << k_funcinfo << "itemFromString failed for  " << s << endl;
			continue;
		}
		ignoreList[mask] = item;
	}

}

void Server::setIgnoreList(IgnoreList l)
{
	ignoreList = l;
}

Server *Server::getServer()
{
	return this;
}

void Server::saveIgnoreList()
{
    KConfig * config = kapp->config();
    QStringList sl;
    IgnoreList::Iterator it;
    for (it = ignoreList.begin(); it != ignoreList.end(); ++it)
	    sl.append(it.key() + it.data().toString());
    config->setGroup(group);
    config->writeEntry("ignoreList", sl);
}

void Server::startConnect()
{
	if (isConnected())  {
		logInfoMsg(i18n("Already connected to %1").arg(host));
		return;
	}
	createSocket();
	connectionTimeoutTimer.start(1000 * CONNECT_TIMEOUT, true);
	resolver.run();
	connecting = true;
}

// Called when socket emits the signal lookupFinished
void Server::lookupFinished()
{
	socket->startAsyncConnect();
}

void Server::handleSSLHandshake(bool res)
{
	if (res) {
		logInfoMsg(i18n("SSL Handshake successful"));
		makeConnect();
	} else {
		logErrorMsg(i18n("SSL Handshake failed"));
	}
}

void Server::handleConnectionSuccess()
{
	connectionTimeoutTimer.stop();
#ifdef HAVE_SSL
	if (ssl) {
		KExtendedSSLSocket * s = (KExtendedSSLSocket *) socket;
		s->startHandshake();
	} else
#endif
        makeConnect();
}

void Server::handleConnectTimeout()
{
	logErrorMsg(i18n("Connection timeout"));
	resolver.terminate();
	socket->cancelAsyncConnect();
	if (socket) {
		delete socket;
		socket = NULL;
	}
	doReconnect();
}

void Server::clearAllChans()
{
	QDictIterator<Channel> itt( channels ); // See QDictIterator
	for( ; itt.current(); ++itt ) {
		Channel *c =  itt.current();
		c->clear();
	}
}

void Server::clearAll()
{
	QDictIterator<Nick> it(nicks); // See QDictIterator
	for( ; it.current(); ++it ) {
		Nick *nick = it.current();
		nick->clear();
	}

	clearAllChans();
}
