/*
 * $Id: server.h,v 1.55 2004/03/07 15:16:07 blackmor Exp $
 *
 * This file is part of the kwirc project
 * Copyright (C) 2003 nohar, Julien Perrot
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * See the file "COPYING" for the exact licensing terms.
 */
#ifndef SERVER_H
#define SERVER_H
#include <kextsock.h>
#include <qdict.h>
#include <qmap.h>
#include <qtimer.h>
#include <qdatetime.h>
#include "serverelem.h"
#include "nick.h"
#include "channel.h"
#include <qobject.h>
#include <qguardedptr.h>
#include "config.h"
#include "ircresolver.h"
#include "serversettings.h"

class Kwirc;

class IgnoreListItem
{
public:
	IgnoreListItem(bool chan = false, bool priv = false,
			bool notice = false, bool invite = false);
	bool chan, priv, notice, invite;
	QString toString();
	QString toPrettyString();
	static IgnoreListItem itemFromString(const QString &, bool *r);
};

typedef QMap<QString,IgnoreListItem> IgnoreList;

class Server : public ServerElem
{
	Q_OBJECT
private:
	char *charBuffer;
	static const Q_ULONG CHAR_BUFFER_SIZE;
	static const Q_ULONG IRC_MAX_MESSAGE_LENGTH;
	static const Q_ULONG CHECK_INTERVAL;
	static const QString LAG_TAG;
	static const Q_ULONG MAX_RETRY;
	static const Q_ULONG DELAY_BEFORE_RETRY;
	static const Q_ULONG CONNECT_TIMEOUT;

	QString group;
	IgnoreList ignoreList;

	QString username_tosend;
	QString pass_tosend;
	QString nick_tosend;

	QString cur_string;

	KExtendedSocket *socket;

	QString host;
	Q_UINT16 port;
	bool ssl;

	int maxModeCount;
	QString chanTypes;
	QString nickPrefixes;

	QDict<Nick> nicks;       // all nicks "known by this server

	QDict<Channel> channels;
	QStringList channelsStr;

	QString serverString;
	QString name;
	Nick *yourNick;
	IRCResolver resolver;

	QTimer *floodTimer;
	unsigned int sentBytesCount;
	static const unsigned int MAX_SENT_BYTES_COUNT = 2 * 80;
	QStringList tbsStrings;
	bool floodMode;
	void putservFloodProtection(const QString &str);
	void realPutserv(const QString &str);

	QTime notifySent;
	QTimer notifyTimer;
	QTimer notifyCheckTimer;
	QTimer connectionTimeoutTimer;

	int currentLag;

	uint retryCounter;
	bool lookupDone;

	QString autojoin;
	QString autojoinkeys;

	bool connecting;

	void do001(QString &, QString &);
	void do005(QString &, QString &);
	void do275(QString &, QString &);
	void do301(QString &, QString &);
	void do401(QString &, QString &);
	void do305(QString &, QString &);
	void do306(QString &, QString &);
	void do307(QString &, QString &);
	void do311(QString &, QString &);
	void do312(QString &, QString &);
	void do313(QString &, QString &);
	void do315(QString &, QString &);
	void do317(QString &, QString &);
	void do318(QString &, QString &);
	void do319(QString &, QString &);
	void do324(QString &, QString &);
	void do332(QString &, QString &);
	void do333(QString &, QString &);
	void do352(QString &, QString &);
	void do353(QString &, QString &);
	void do366(QString &, QString &);
	void do367(QString &, QString &);
	void do368(QString &, QString &);
	void do376();
	void do433(QString &, QString &);
	void doElse(QString &,QString &,QString &);
	void doError(QString &,QString &);
	void doInvite(QString &,QString &);
	void doJoin(QString &,QString &);
	void doKick(QString &,QString &);
	void doKill(QString &,QString &);
	void doMode(QString &,QString &);
	void doNick(QString &,QString &);
	void doNotice(QString &,QString &);
	void doServerNotice(const QString &str);
	void doPart(QString &,QString &);
	void doPrivate(Nick *nick, QString &message);
	void doPrivateNotice(Nick *nick, QString &message);
	void doPrivmsg(QString &,QString &);
	void doPublic(Nick *nick, Channel *channel, QString &message);
	void doPublicNotice(Nick *nick, Channel *channel, QString &message);
	void doQuit(QString &,QString &);
	void doTopic(QString &,QString &);
	void doWhois(QString &w, QString &what);

	bool ctcpMessage(QString &message);

	void wakeUp();
	void waitAWhile();
	void insertNick(Nick *n);
	void removeNick(Nick *n);

	void createSocket();

	void logMsg(KwircLogType level,  const QString &msg);
	void logErrorMsg(const QString &msg);
	void logInfoMsg(const QString &msg);

	bool isIgnored(const QString &ircid, bool chan, bool priv, bool notice,
			bool invite);
	void setIgnoreListFromStringList(QStringList l);
public:
	static const Q_ULONG MAX_LAG;

	Server(const QString &host,  Q_UINT16 port, const QString &label,
	       const QString &username, const QString &pass, bool ssl,
	       const QString &nick, const QString &configGroup,
	       const QStringList &ignoreList);

	virtual ~Server();

	// XXX whatch me! create = true!
	Nick *nickFromString(const QString &what, bool create = true);
	Channel *channel(const QString &what, bool create = false);
	void removeChannel(Channel *c);

	Nick *getYourNick();

	void putserv(const QString &);
	void getserv(const QString &);

	void sendAway(const QString &a);
	void sendNick(const QString &);
	void sendJoin(const QString &, const QString &keys = 0);
	void sendPart(const QString &chan, const QString &msg = QString::null);
	void sendUser(const QString &, const QString &, const QString &,
			const QString &);
	void sendPass(const QString &p);
	void sendPrivmsg(const QString &, const QString &);
	void sendAction(const QString &to, const QString &msg);
	void sendWho(const QString &a);
	void sendWhois(const QString &a);
	void sendOper(const QString &, const QString &);
	void sendQuit(const QString &message = QString::null);
	void sendMode(const QString &chan, const QString &mode,
			const QString &arg);

	void sendCtcp(const QString &what, const QString &dest);
	void sendCtcpNotice(const QString &dest, const QString &what);

	void sendTopic(const QString &channel, const QString &topic);

	bool changeServer(const QString &ahost, Q_UINT16 aport,
			const QString &username, const QString &pass, bool ssl,
			const QString &nick);

	bool isQueryOpen(Nick *n);
	void deleteNick(Nick *n);

	void setAutoJoin(QString aj, QString ajk);
	bool isConnected();

	void addIgnoreMask(const QString &mask, bool chan, bool priv,
			bool notice, bool invite);
	void setIgnoreMask(const QString &mask, bool chan, bool priv,
			bool notice, bool invite);
	void removeIgnoreMask(const QString &mask);
	const IgnoreList &getIgnoreList();
	void setIgnoreList(IgnoreList l);

	const QDict<Channel> &getChannels();
	const QDict<Nick> &getNicks();
	virtual Server *getServer();
	virtual bool eventFilter(QObject* parent,QEvent* event);
	void cleanup();
	void clearAllChans();
	void clearAll();
public slots:
	void makeConnect();
	void serverTalks();  // socket is talking
	void serverCloses();  // socket is talking
	//void timerDone();	// idler
	void slotQuery(const QString &nick);
	void startNotifyTimer(int sec);
	void startConnect();
	void doReconnect();
	void saveIgnoreList();
private slots:
	void socketClosed(int);
	void notifyTimeout();
	void notifyCheckTimeout();
	void handleConnectionFailure(int);
	void lookupFinished();
	void floodTimerTrigger();
	void handleConnectionSuccess();
	void handleSSLHandshake(bool);
	void handleConnectTimeout();
signals:
	void newQuery(Nick *n);
	void newChannel(Channel *c);
	void delChannel(Channel *c);
	void lag(int);
	void logAll(KwircLogType, const QString &, const QString &,
			const QString &);
};

#endif
